@@ -1,7 +1,6 @@
-use std::ffi::c_void;
 use std::fmt::Debug;
 
-use super::{InputStore, InputStoreKer, OutputStore, OutputStoreKer, PackedStore};
+use super::{InputStore, OutputStore, OutputStoreKer, PackedStore};
 use tract_data::internal::*;
 
 #[repr(usize)]
@@ -49,6 +48,20 @@ pub enum FusedSpec<'t> {
     AddMatMul { k: usize, a: PackedStore, b: InputStore },
 }
 
+impl<'t> FusedSpec<'t> {
+    pub fn prefer_col_outer(&self) -> bool {
+        if let FusedSpec::AddMatMul { b, .. } = self {
+            match &b {
+                &InputStore::Packed { .. } => false,
+                &InputStore::VirtualPacking { .. } => true,
+                &InputStore::LatePacking { .. } => true,
+            }
+        } else {
+            false
+        }
+    }
+}
+
 // Careful here, the jump_to comments are used by the build script.
 #[repr(C, usize)]
 #[derive(PartialEq, Copy, Clone, Debug)]
@@ -84,19 +97,18 @@ pub enum FusedKerSpec<TI: Copy> {
     Store(OutputStoreKer),                      // jump_to:store
 
     // jump_to:add_mat_mul
-    AddMatMul { k: usize, pa: *const c_void, pb: *const InputStoreKer, cpu_variant: usize },
+    AddMatMul { k: usize, pa: *const u8, pb: *const u8, cpu_variant: usize },
 }
 
 #[cfg(test)]
 #[macro_use]
 pub mod test {
-    use super::*;
     use crate::frame::mmm::storage::*;
     use crate::frame::mmm::*;
     use crate::generic::ScaleShiftAndRound;
-    use num_traits::{AsPrimitive, Bounded, Zero};
+    use num_traits::{AsPrimitive, Bounded};
     use proptest::prelude::*;
-    use std::ops::{Add, Mul, Sub};
+    use tract_data::internal::*;
 
     #[test]
     fn check_non_linear_enum_size() {
@@ -113,11 +125,12 @@ pub mod test {
 
     #[macro_export]
     macro_rules! mmm_kernel_fuse_tests {
-        ($cond:expr, $ker:ty, $tc:ty, $ti: ty) => {
+        ($cond:expr, $ker:ident, $tc:ty, $ti: ty) => {
             mod fuse {
                 #[allow(unused_imports)]
                 use crate::frame::mmm::fuse::test;
                 use crate::frame::mmm::fuse::test::tile;
+                use super::super::$ker;
 
                 #[test]
                 fn return_zeros() {
@@ -252,7 +265,7 @@ pub mod test {
 
     #[macro_export]
     macro_rules! qmmm_kernel_fuse_tests {
-        ($cond:expr, $ker:ty, $ta:ty, $tb:ty, $tc:ty, $ti: ty) => {
+        ($cond:expr, $ker:ident, $ta:ty, $tb:ty, $tc:ty, $ti: ty) => {
             mod fuseq {
                 use crate::frame::mmm::fuse::RoundingPolicy;
                 #[allow(unused_imports)]
@@ -260,6 +273,7 @@ pub mod test {
                 use crate::frame::mmm::fuse::test::QScaleProblem;
                 use crate::frame::mmm::kernel::MatMatMulKer;
                 use proptest::prelude::*;
+                use super::super::$ker;
 
                 #[test]
                 fn return_q_scale_0() {
@@ -329,30 +343,30 @@ pub mod test {
                                     let len = (<$ker>::mr() * <$ker>::nr()) as i64;
                                     let v = (0..len).map(|i| (i - len / 2) as $tc).collect();
                                     QScaleProblem::<$ker, $tc, $ti>::new(v, 1<<30, 2, RoundingPolicy::$policy).run()
+                                }
                             }
                         }
                     }
                 }
-            }
 
-            test_q_scale!(Zero);
-            test_q_scale!(Away);
-            test_q_scale!(MinusInf);
-            test_q_scale!(PlusInf);
-            test_q_scale!(Even);
-            test_q_scale!(Odd);
+                test_q_scale!(Zero);
+                test_q_scale!(Away);
+                test_q_scale!(MinusInf);
+                test_q_scale!(PlusInf);
+                test_q_scale!(Even);
+                test_q_scale!(Odd);
 
-            proptest::proptest! {
-                #[test]
-                fn return_q_scale_prop(pb in any::<QScaleProblem<$ker, $tc, $ti>>()) {
-                    if $cond {
-                        pb.run()
+                proptest::proptest! {
+                    #[test]
+                    fn return_q_scale_prop(pb in any::<QScaleProblem<$ker, $tc, $ti>>()) {
+                        if $cond {
+                            pb.run()
+                        }
                     }
                 }
             }
-        }
-    };
-}
+        };
+    }
 
     pub fn mmm_stride_storage<T: Copy>(v: &[T], rsc: usize) -> OutputStoreKer {
         OutputStoreKer {
@@ -363,11 +377,12 @@ pub mod test {
         }
     }
 
+    use crate::LADatum;
     pub fn return_zeros<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Copy + Debug + Bounded + Zero + PartialEq,
-        TI: Copy + Debug,
+        TC: LADatum,
+        TI: LADatum + Bounded + PartialEq,
     {
         let mut v = vec![TC::max_value(); K::mr() * K::nr()];
         let c = mmm_stride_storage(&mut v, K::nr());
@@ -381,8 +396,8 @@ pub mod test {
     pub fn fused_ops<K, TC, TI, E>(c: &[TC], ops: &[FusedKerSpec<TI>], expect: E)
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC>,
+        TC: Datum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         E: Fn(usize, usize, TI) -> TI,
     {
         assert!(c.len() == K::mr() * K::nr());
@@ -417,8 +432,8 @@ pub mod test {
     pub fn return_c<K, TC, TI>(v: &[TC])
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Zero + Add<Output = TI> + Sub<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         fused_ops::<K, TC, TI, _>(&*v, &[], |_, _, c| c + 1.as_() - 1.as_())
@@ -427,8 +442,8 @@ pub mod test {
     pub fn return_c_plus_d<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Add<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -444,8 +459,8 @@ pub mod test {
     pub fn return_c_min_row<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + std::cmp::PartialOrd,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -463,8 +478,8 @@ pub mod test {
     pub fn return_c_max_row<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + std::cmp::PartialOrd,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -482,8 +497,8 @@ pub mod test {
     pub fn return_c_add_row<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Add<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -497,8 +512,8 @@ pub mod test {
     pub fn return_c_mul_row<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Mul<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -512,8 +527,8 @@ pub mod test {
     pub fn return_c_sub_row<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Sub<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -527,8 +542,8 @@ pub mod test {
     pub fn return_c_subf_row<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Sub<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -542,8 +557,8 @@ pub mod test {
     pub fn return_c_add_col<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Add<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -557,8 +572,8 @@ pub mod test {
     pub fn return_c_mul_col<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Mul<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -572,8 +587,8 @@ pub mod test {
     pub fn return_c_add_row_col_product<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Add<Output = TI> + Mul<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -590,8 +605,8 @@ pub mod test {
     pub fn return_c_scalar_min<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + std::cmp::PartialOrd,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -608,8 +623,8 @@ pub mod test {
     pub fn return_c_scalar_max<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + std::cmp::PartialOrd,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -626,8 +641,8 @@ pub mod test {
     pub fn return_c_scalar_add<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Add<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -638,8 +653,8 @@ pub mod test {
     pub fn return_c_scalar_mul<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Mul<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -650,8 +665,8 @@ pub mod test {
     pub fn return_c_scalar_sub<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Sub<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -663,8 +678,8 @@ pub mod test {
     pub fn return_c_scalar_subf<K, TC, TI>()
     where
         K: MatMatMulKer<TI>,
-        TC: Datum + Copy + AsPrimitive<TI>,
-        TI: Datum + Copy + AsPrimitive<TC> + Sub<Output = TI>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
     {
         let len = K::mr() * K::nr();
@@ -677,8 +692,8 @@ pub mod test {
     pub struct QScaleProblem<K, TC, TI>
     where
         K: MatMatMulKer<TI>,
-        TC: Copy + Debug + 'static,
-        TI: Copy + Debug,
+        TC: LADatum,
+        TI: LADatum + AsPrimitive<TC>,
         i64: AsPrimitive<TC>,
     {
         pub c: Vec<TC>,
@@ -691,8 +706,8 @@ pub mod test {
     impl<K, TC, TI> Arbitrary for QScaleProblem<K, TC, TI>
     where
         K: MatMatMulKer<TI>,
-        TC: Copy + Debug + Arbitrary + 'static,
-        TI: Copy + Debug,
+        TC: LADatum + Arbitrary,
+        TI: LADatum + AsPrimitive<TC>,
         i64: AsPrimitive<TC>,
     {
         type Parameters = ();
@@ -727,8 +742,8 @@ pub mod test {
     impl<K, TC, TI> QScaleProblem<K, TC, TI>
     where
         K: MatMatMulKer<TI>,
-        TC: Copy + Datum + 'static + AsPrimitive<TI> + PartialEq,
-        TI: Copy + Datum + 'static + AsPrimitive<i64> + ScaleShiftAndRound + AsPrimitive<TC>,
+        TC: LADatum + AsPrimitive<TI>,
+        TI: LADatum + AsPrimitive<TC> + ScaleShiftAndRound + AsPrimitive<i64>,
         usize: AsPrimitive<TC> + AsPrimitive<TI>,
         i64: AsPrimitive<TC>,
     {
@@ -744,8 +759,8 @@ pub mod test {
     pub fn tile<K, TC, TI>() -> BoxedStrategy<Vec<TC>>
     where
         K: MatMatMulKer<TI>,
-        TC: Copy + Debug + 'static,
-        TI: Copy + Debug,
+        TC: LADatum,
+        TI: LADatum + AsPrimitive<TC>,
         i8: AsPrimitive<TC>,
     {
         let len = K::mr() * K::nr();
